@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2013-2021 Canonical, Ltd.
- * Copyright (C) 2022-2023 Colin Ian King.
+ * Copyright (C) 2022-2024 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -18,6 +18,7 @@
  *
  */
 #include "stress-ng.h"
+#include "core-attribute.h"
 
 static const stress_help_t help[] = {
 	{ NULL,	"chown N",	"start N workers thrashing chown file ownership" },
@@ -25,6 +26,35 @@ static const stress_help_t help[] = {
 	{ NULL,	NULL,		NULL }
 };
 
+static int PURE OPTIMIZE3 stress_chown_check(const int ret)
+{
+	static const int ignore_errors[] = {
+#if defined(ENOENT)
+		ENOENT,
+#endif
+#if defined(ENOTDIR)
+		ENOTDIR,
+#endif
+#if defined(ENOSYS)
+		ENOSYS,
+#endif
+#if defined(EPERM)
+		EPERM,
+#endif
+	};
+
+	size_t i;
+
+	if (!ret)
+		return 0;
+
+	for (i = 0; i < SIZEOF_ARRAY(ignore_errors); i++) {
+		if (errno == ignore_errors[i])
+			return 0;
+	}
+	return -1;
+}
+
 /*
  *  do_fchown()
  *	set ownership different ways
@@ -36,30 +66,34 @@ static int do_fchown(
 	const uid_t uid,
 	const gid_t gid)
 {
-	int tmp;
+	int tmp, ret;
 
-	if (fchown(fd, uid, gid) < 0)
+	if (stress_chown_check(fchown(fd, uid, gid) < 0))
 		return -errno;
-	if (fchown(fd, (uid_t)-1, gid) < 0)
+	if (stress_chown_check(fchown(fd, (uid_t)-1, gid) < 0))
 		return -errno;
-	if (fchown(fd, uid, (gid_t)-1) < 0)
+	if (stress_chown_check(fchown(fd, uid, (gid_t)-1) < 0))
 		return -errno;
-	if (fchown(fd, (uid_t)-1, (gid_t)-1) < 0)
+	if (stress_chown_check(fchown(fd, (uid_t)-1, (gid_t)-1) < 0))
 		return -errno;
 
 	if (cap_chown)
 		return 0;
-	if (fchown(fd, (uid_t)0, (gid_t)0) == 0)
+	ret = fchown(fd, (uid_t)0, (gid_t)0);
+	if (ret == 0)
 		goto restore;
-	if (errno != EPERM)
+	if (stress_chown_check(ret) < 0)
 		goto restore;
-	if ((fchown(fd, (uid_t)-1, (gid_t)0) == 0) && (errno != EPERM))
+
+	ret = fchown(fd, (uid_t)-1, (gid_t)0);
+	if (ret == 0)
 		goto restore;
-	if (errno != EPERM)
+	if (stress_chown_check(ret) < 0)
 		goto restore;
-	if (fchown(fd, (uid_t)0, (gid_t)-1) == 0)
+	ret = fchown(fd, (uid_t)0, (gid_t)-1);
+	if (ret == 0)
 		goto restore;
-	if (errno != EPERM)
+	if (stress_chown_check(ret) < 0)
 		goto restore;
 
 	/*
@@ -87,32 +121,34 @@ static int do_chown(
 	const uid_t uid,
 	const gid_t gid)
 {
-	int tmp;
+	int tmp, ret;
 
-	if (chown_func(filename, uid, gid) < 0)
+	if (stress_chown_check(chown_func(filename, uid, gid)) < 0)
 		return -errno;
-	if (chown_func(filename, (uid_t)-1, gid) < 0)
+	if (stress_chown_check(chown_func(filename, (uid_t)-1, gid)) < 0)
 		return -errno;
-	if (chown_func(filename, uid, (gid_t)-1) < 0)
+	if (stress_chown_check(chown_func(filename, uid, (gid_t)-1)) < 0)
 		return -errno;
-	if (chown_func(filename, (uid_t)-1, (gid_t)-1) < 0)
+	if (stress_chown_check(chown_func(filename, (uid_t)-1, (gid_t)-1)) < 0)
 		return -errno;
 
 	if (cap_chown)
 		return 0;
-	if (chown_func(filename, (uid_t)0, (gid_t)0) == 0)
+	ret = chown_func(filename, (uid_t)0, (gid_t)0);
+	if (ret == 0)
 		goto restore;
-	if (errno != EPERM)
+	if (stress_chown_check(ret))
 		goto restore;
-	if ((chown_func(filename, (uid_t)-1, (gid_t)0) == 0) && (errno != EPERM))
+	ret = chown_func(filename, (uid_t)-1, (gid_t)0);
+	if (ret == 0)
 		goto restore;
-	if (errno != EPERM)
+	if (stress_chown_check(ret))
 		goto restore;
-	if (chown_func(filename, (uid_t)0, (gid_t)-1) == 0)
+	ret = chown_func(filename, (uid_t)0, (gid_t)-1);
+	if (ret == 0)
 		goto restore;
-	if (errno != EPERM)
+	if (stress_chown_check(ret))
 		goto restore;
-
 	return 0;
 
 restore:
@@ -126,7 +162,7 @@ restore:
  *  stress_chown
  *	stress chown
  */
-static int stress_chown(const stress_args_t *args)
+static int stress_chown(stress_args_t *args)
 {
 	const pid_t ppid = getppid();
 	int fd = -1, rc = EXIT_FAILURE, retries = 0;
@@ -164,7 +200,7 @@ static int stress_chown(const stress_args_t *args)
 	} else {
 		/* Other instances must try to open the file */
 		for (;;) {
-			if ((fd = open(filename, O_RDWR, S_IRUSR | S_IWUSR)) > - 1)
+			if ((fd = open(filename, O_RDWR)) > - 1)
 				break;
 
 #if defined(__NetBSD__)
@@ -175,16 +211,17 @@ static int stress_chown(const stress_args_t *args)
 			(void)shim_usleep(100000);
 #endif
 			/* Timed out, then give up */
-			if (!keep_stressing_flag()) {
+			if (!stress_continue_flag()) {
 				rc = EXIT_SUCCESS;
 				goto tidy;
 			}
-			if (++retries >= 100) {
-				pr_err("%s: chown: file %s took %d "
+			if (++retries >= 1000) {
+				pr_inf("%s: chown: file %s took %d "
 					"retries to open and gave up "
 					"(instance %" PRIu32 ")%s\n",
 					args->name, filename, retries, args->instance,
-					stress_fs_type(filename));
+					stress_get_fs_type(filename));
+				rc = EXIT_NO_RESOURCE;
 				goto tidy;
 			}
 		}
@@ -199,41 +236,23 @@ static int stress_chown(const stress_args_t *args)
 		if ((ret < 0) && (ret != -EPERM))
 			pr_fail("%s: fchown failed, errno=%d (%s)%s\n",
 				args->name, errno, strerror(errno),
-				stress_fs_type(filename));
+				stress_get_fs_type(filename));
 
 		ret = do_chown(chown, filename, cap_chown, uid, gid);
-		if (ret < 0) {
-			if ((ret == -ENOENT) || (ret == -ENOTDIR)) {
-				/*
-				 * File was removed during test by
-				 * another worker
-				 */
-				rc = EXIT_SUCCESS;
-				goto tidy;
-			}
-			if (ret != -EPERM)
-				pr_fail("%s: chown %s failed, errno=%d (%s)%s\n",
-					args->name, filename, errno, strerror(errno),
-					stress_fs_type(filename));
+		if ((ret < 0) && (ret != -EPERM)) {
+			pr_fail("%s: chown %s failed, errno=%d (%s)%s\n",
+				args->name, filename, errno, strerror(errno),
+				stress_get_fs_type(filename));
 		}
 		ret = do_chown(lchown, filename, cap_chown, uid, gid);
-		if (ret < 0) {
-			if ((ret == -ENOENT) || (ret == -ENOTDIR)) {
-				/*
-				 * File was removed during test by
-				 * another worker
-				 */
-				rc = EXIT_SUCCESS;
-				goto tidy;
-			}
-			if (ret != -EPERM)
-				pr_fail("%s: lchown %s failed, errno=%d (%s)%s\n",
-					args->name, filename, errno, strerror(errno),
-					stress_fs_type(filename));
+		if ((ret < 0) && (ret != -EPERM)) {
+			pr_fail("%s: lchown %s failed, errno=%d (%s)%s\n",
+				args->name, filename, errno, strerror(errno),
+				stress_get_fs_type(filename));
 		}
 		(void)shim_fsync(fd);
-		inc_counter(args);
-	} while (keep_stressing(args));
+		stress_bogo_inc(args);
+	} while (stress_continue(args));
 
 	rc = EXIT_SUCCESS;
 tidy:
