@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2013-2021 Canonical, Ltd.
- * Copyright (C) 2022-2023 Colin Ian King.
+ * Copyright (C) 2022-2024 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -18,6 +18,7 @@
  *
  */
 #include "stress-ng.h"
+#include "core-builtin.h"
 
 #if defined(HAVE_SYS_SELECT_H)
 #include <sys/select.h>
@@ -95,58 +96,89 @@ static int stress_set_dentry_order(const char *opt)
 	return -1;
 }
 
+/*
+ *  stress_dentry_unlink_file()
+ *	unlink a file. if verify mode is enabled, read and check
+ *	contents to make sure it matches the expected gray code
+ */
+static void stress_dentry_unlink_file(
+	stress_args_t *args,
+	const uint64_t gray_code,
+	const bool verify,
+	uint64_t *read_errors)
+{
+	char path[PATH_MAX];
+
+	stress_temp_filename_args(args, path, sizeof(path), gray_code * 2);
+	if (verify) {
+		int fd;
+		uint64_t val;
+
+		fd = open(path, O_RDONLY);
+		if (fd >= 0) {
+			ssize_t rret;
+
+			rret = read(fd, &val, sizeof(val));
+			if ((rret == sizeof(val)) && (val != gray_code)) {
+				pr_inf("err: %" PRIx64 " vs %" PRIx64 "\n",
+					val, gray_code);
+				(*read_errors)++;
+			}
+			(void)close(fd);
+		}
+	}
+	(void)shim_unlink(path);
+}
+
 /*
  *  stress_dentry_unlink()
  *	remove all dentries
  */
 static void stress_dentry_unlink(
-	const stress_args_t *args,
+	stress_args_t *args,
 	const uint64_t n,
-	const uint8_t dentry_order)
+	const uint8_t dentry_order,
+	const bool verify)
 {
 	uint64_t i, j;
 	uint64_t prime;
+	uint64_t read_errors = 0ULL;
 	const uint8_t ord = (dentry_order == ORDER_RANDOM) ?
 				stress_mwc8modn(3) : dentry_order;
 
 	switch (ord) {
 	case ORDER_REVERSE:
 		for (i = 0; i < n; i++) {
-			char path[PATH_MAX];
 			uint64_t gray_code;
 
 			j = (n - 1) - i;
 			gray_code = (j >> 1) ^ j;
-
-			stress_temp_filename_args(args,
-				path, sizeof(path), gray_code * 2);
-			(void)shim_unlink(path);
+			stress_dentry_unlink_file(args, gray_code, verify, &read_errors);
 		}
 		break;
 	case ORDER_STRIDE:
-		prime = stress_get_prime64(n);
+		prime = stress_get_next_prime64(n);
 		for (i = 0, j = prime; i < n; i++, j += prime) {
-			char path[PATH_MAX];
 			const uint64_t k = j % n;
 			const uint64_t gray_code = (k >> 1) ^ k;
 
-			stress_temp_filename_args(args,
-				path, sizeof(path), gray_code * 2);
-			(void)shim_unlink(path);
+			stress_dentry_unlink_file(args, gray_code, verify, &read_errors);
 		}
 		break;
 	case ORDER_FORWARD:
 	default:
 		for (i = 0; i < n; i++) {
-			char path[PATH_MAX];
 			const uint64_t gray_code = (i >> 1) ^ i;
 
-			stress_temp_filename_args(args,
-				path, sizeof(path), gray_code * 2);
-			(void)shim_unlink(path);
+			stress_dentry_unlink_file(args, gray_code, verify, &read_errors);
 		}
 		break;
 	}
+
+	if (read_errors > 0) {
+		pr_fail("%s: %" PRIu64 " files did not contain the expected graycode check data\n",
+			args->name, read_errors);
+	}
 }
 
 /*
@@ -197,11 +229,11 @@ static void stress_dentry_misc(const char *path)
 
 #if defined(HAVE_UTIME_H)
 	(void)utime(path, NULL);
-	(void)memset(&utim, 0, sizeof(utim));
+	(void)shim_memset(&utim, 0, sizeof(utim));
 	(void)utime(path, &utim);
 #endif
 
-	VOID_RET(int, fstat(fd, &statbuf));
+	VOID_RET(int, shim_fstat(fd, &statbuf));
 
 	/* Not really legal */
 	VOID_RET(off_t, lseek(fd, 0, SEEK_END));
@@ -218,7 +250,7 @@ static void stress_dentry_misc(const char *path)
 	VOID_RET(int, shim_fallocate(fd, 0, (off_t)0, statbuf.st_size));
 
 	/* mmap */
-	ptr = mmap(NULL, 4096, PROT_READ, MAP_ANONYMOUS | MAP_PRIVATE, fd, 0);
+	ptr = stress_mmap_populate(NULL, 4096, PROT_READ, MAP_ANONYMOUS | MAP_PRIVATE, fd, 0);
 	if (ptr != MAP_FAILED)
 		(void)munmap(ptr, 4096);
 
@@ -306,7 +338,7 @@ static void stress_dentry_misc(const char *path)
  *	on a gray-coded value multiplied by two.
  *	Even numbered files exist, odd don't exist.
  */
-static int stress_dentry(const stress_args_t *args)
+static int stress_dentry(stress_args_t *args)
 {
 	int ret;
 	uint64_t dentries = DEFAULT_DENTRIES;
@@ -319,6 +351,7 @@ static int stress_dentry(const stress_args_t *args)
 	double bogus_access_duration = 0.0, bogus_access_count = 0.0;
 	double bogus_unlink_duration = 0.0, bogus_unlink_count = 0.0;
 	double rate;
+	const bool verify = !!(g_opt_flags & OPT_FLAGS_VERIFY);
 
 	if (!stress_get_setting("dentries", &dentries)) {
 		if (g_opt_flags & OPT_FLAGS_MAXIMIZE)
@@ -346,7 +379,7 @@ static int stress_dentry(const stress_args_t *args)
 			int fd;
 			double t;
 
-			if (!keep_stressing(args))
+			if (!stress_continue(args))
 				goto abort;
 
 			stress_temp_filename_args(args,
@@ -363,9 +396,18 @@ static int stress_dentry(const stress_args_t *args)
 			}
 			creat_duration += stress_time_now() - t;
 			creat_count += 1.0;
-			(void)close(fd);
 
-			inc_counter(args);
+			if (verify) {
+				ssize_t wret;
+
+				wret = write(fd, &gray_code, sizeof(gray_code));
+				if (wret < 0) {
+					(void)close(fd);
+					break;
+				}
+			}
+			(void)close(fd);
+			stress_bogo_inc(args);
 		}
 
 		stress_dentry_misc(dir_path);
@@ -379,7 +421,7 @@ static int stress_dentry(const stress_args_t *args)
 			const uint64_t gray_code = (i >> 1) ^ i;
 			double t;
 
-			if (!keep_stressing(args))
+			if (!stress_continue(args))
 				goto abort;
 
 			/* The following should succeed */
@@ -421,12 +463,12 @@ static int stress_dentry(const stress_args_t *args)
 		/*
 		 *  And remove
 		 */
-		stress_dentry_unlink(args, n, dentry_order);
+		stress_dentry_unlink(args, n, dentry_order, verify);
 		stress_dentry_misc(dir_path);
 
-		if (!keep_stressing_flag())
+		if (!stress_continue_flag())
 			break;
-	} while (keep_stressing(args));
+	} while (stress_continue(args));
 
 abort:
 	stress_dentry_state(&nr_dentry2);
@@ -438,18 +480,20 @@ abort:
 	stress_set_proc_state(args->name, STRESS_STATE_DEINIT);
 
 	rate = (creat_count > 0.0) ? (double)creat_duration / creat_count : 0.0;
-	stress_metrics_set(args, 0, "nanosecs per file creation", rate * STRESS_DBL_NANOSECOND);
+	stress_metrics_set(args, 0, "nanosecs per file creation",
+		rate * STRESS_DBL_NANOSECOND, STRESS_HARMONIC_MEAN);
 	rate = (access_count > 0.0) ? (double)access_duration / access_count : 0.0;
-	stress_metrics_set(args, 1, "nanosecs per file access", rate * STRESS_DBL_NANOSECOND);
+	stress_metrics_set(args, 1, "nanosecs per file access",
+		rate * STRESS_DBL_NANOSECOND, STRESS_HARMONIC_MEAN);
 	rate = (bogus_access_count > 0.0) ? (double)bogus_access_duration / bogus_access_count : 0.0;
-	stress_metrics_set(args, 2, "nanosecs per bogus file access", rate * STRESS_DBL_NANOSECOND);
+	stress_metrics_set(args, 2, "nanosecs per bogus file access",
+		rate * STRESS_DBL_NANOSECOND, STRESS_HARMONIC_MEAN);
 	rate = (bogus_unlink_count > 0.0) ? (double)bogus_unlink_duration / bogus_unlink_count : 0.0;
-	stress_metrics_set(args, 3, "nanosecs per bogus file unlink", rate * STRESS_DBL_NANOSECOND);
+	stress_metrics_set(args, 3, "nanosecs per bogus file unlink",
+		rate * STRESS_DBL_NANOSECOND, STRESS_HARMONIC_MEAN);
 
 	/* force unlink of all files */
-	pr_tidy("%s: removing %" PRIu64 " entries\n",
-		args->name, dentries);
-	stress_dentry_unlink(args, dentries, dentry_order);
+	stress_dentry_unlink(args, dentries, dentry_order, verify);
 	(void)stress_temp_dir_rm_args(args);
 
 	return EXIT_SUCCESS;
@@ -465,5 +509,6 @@ stressor_info_t stress_dentry_info = {
 	.stressor = stress_dentry,
 	.class = CLASS_FILESYSTEM | CLASS_OS,
 	.opt_set_funcs = opt_set_funcs,
+	.verify = VERIFY_OPTIONAL,
 	.help = help
 };
