@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2021-2023 Colin Ian King
+ * Copyright (C) 2021-2024 Colin Ian King
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -17,6 +17,9 @@
  *
  */
 #include "stress-ng.h"
+#include "core-builtin.h"
+#include "core-cpu-cache.h"
+#include "core-pragma.h"
 #include "core-put.h"
 
 #define STRESS_RANDLIST_DEFAULT_ITEMS	(100000)
@@ -37,6 +40,7 @@ static const stress_help_t help[] = {
 
 typedef struct stress_randlist_item {
 	struct stress_randlist_item *next;
+	uint8_t dataval;
 	uint8_t alloc_type:1;
 	uint8_t data[];
 } stress_randlist_item_t;
@@ -115,17 +119,60 @@ static void stress_randlist_free_ptrs(
 	free(ptrs);
 }
 
-static void stress_randlist_enomem(const stress_args_t *args)
+static void stress_randlist_enomem(stress_args_t *args)
 {
 	pr_inf_skip("%s: cannot allocate the list, skipping stressor\n",
 		args->name);
 }
 
+static inline uint8_t OPTIMIZE3 stress_randlist_bad_data(
+	const stress_randlist_item_t *ptr,
+	const size_t randlist_size)
+{
+	register const uint8_t *data = ptr->data;
+	register const uint8_t *end = data + randlist_size;
+	register const uint8_t dataval = ptr->dataval;
+
+PRAGMA_UNROLL_N(8)
+	while (data < end) {
+		if (*(data++) != dataval)
+			return true;
+	}
+	return false;
+}
+
+static inline void OPTIMIZE3 stress_randlist_exercise(
+	stress_args_t *args,
+	stress_randlist_item_t *head,
+	const size_t randlist_size,
+	const bool verify)
+{
+	register stress_randlist_item_t *ptr;
+	uint8_t dataval = stress_mwc8();
+
+	for (ptr = head; ptr; ptr = ptr->next) {
+		shim_builtin_prefetch(ptr->next);
+		ptr->dataval = dataval;
+		(void)shim_memset(ptr->data, dataval, randlist_size);
+		dataval++;
+		if (!stress_continue_flag())
+			break;
+	}
+
+	for (ptr = head; ptr; ptr = ptr->next) {
+		shim_builtin_prefetch(ptr->next);
+		if (!stress_continue_flag())
+			break;
+		if (verify && stress_randlist_bad_data(ptr, randlist_size))
+			pr_fail("%s: data check failure in list object at 0x%p\n", args->name, ptr);
+	}
+}
+
 /*
  *  stress_randlist()
- *	stress a list containing random values 
+ *	stress a list containing random values
  */
-static int stress_randlist(const stress_args_t *args)
+static int stress_randlist(stress_args_t *args)
 {
 	register size_t i;
 	stress_randlist_item_t **ptrs;
@@ -133,6 +180,7 @@ static int stress_randlist(const stress_args_t *args)
 	stress_randlist_item_t *compact_ptr = NULL;
 	bool do_mmap = false;
 	bool randlist_compact = false;
+	const bool verify = !!(g_opt_flags & OPT_FLAGS_VERIFY);
 	size_t randlist_items = STRESS_RANDLIST_DEFAULT_ITEMS;
 	size_t randlist_size = STRESS_RANDLIST_DEFAULT_SIZE;
 	size_t heap_allocs = 0;
@@ -164,7 +212,7 @@ static int stress_randlist(const stress_args_t *args)
 		}
 
 		for (ptr = compact_ptr, i = 0; i < randlist_items; i++) {
-			if (!keep_stressing_flag()) {
+			if (!stress_continue_flag()) {
 				stress_set_proc_state(args->name, STRESS_STATE_DEINIT);
 				stress_randlist_free_ptrs(compact_ptr, ptrs, i, randlist_size);
 				stress_randlist_enomem(args);
@@ -178,15 +226,16 @@ static int stress_randlist(const stress_args_t *args)
 		for (i = 0; i < randlist_items; i++) {
 			const size_t size = sizeof(*ptr) + randlist_size;
 
-			if (!keep_stressing_flag()) {
+			if (!stress_continue_flag()) {
 				stress_set_proc_state(args->name, STRESS_STATE_DEINIT);
 				stress_randlist_free_ptrs(compact_ptr, ptrs, i, randlist_size);
 				return EXIT_SUCCESS;
 			}
 retry:
 			if (do_mmap && (stress_mwc8() < 16)) {
-				ptr = (stress_randlist_item_t *)mmap(NULL, size, PROT_READ | PROT_WRITE,
-					MAP_SHARED | MAP_ANONYMOUS, -1, 0);
+				ptr = (stress_randlist_item_t *)stress_mmap_populate(NULL, size,
+					PROT_READ | PROT_WRITE,
+					MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);
 				if (ptr == MAP_FAILED) {
 					do_mmap = false;
 					goto retry;
@@ -217,7 +266,7 @@ retry:
 		ptrs[i] = ptrs[n];
 		ptrs[n] = ptr;
 
-		if (!keep_stressing_flag()) {
+		if (!stress_continue_flag()) {
 			stress_set_proc_state(args->name, STRESS_STATE_DEINIT);
 			stress_randlist_free_ptrs(compact_ptr, ptrs, i, randlist_size);
 			return EXIT_SUCCESS;
@@ -238,23 +287,9 @@ retry:
 	stress_set_proc_state(args->name, STRESS_STATE_RUN);
 
 	do {
-		for (ptr = head; ptr; ptr = ptr->next) {
-			(void)memset(ptr->data, stress_mwc8(), randlist_size);
-			if (!keep_stressing_flag())
-				break;
-		}
-
-		for (ptr = head; ptr; ptr = ptr->next) {
-			uint32_t sum = 0;
-			for (i = 0; i < randlist_size; i++) {
-				sum += ptr->data[i];
-			}
-			stress_uint32_put(sum);
-			if (!keep_stressing_flag())
-				break;
-		}
-		inc_counter(args);
-	} while (keep_stressing(args));
+		stress_randlist_exercise(args, head, randlist_size, verify);
+		stress_bogo_inc(args);
+	} while (stress_continue(args));
 
 	stress_set_proc_state(args->name, STRESS_STATE_DEINIT);
 
@@ -284,5 +319,6 @@ stressor_info_t stress_randlist_info = {
 	.stressor = stress_randlist,
 	.class = CLASS_MEMORY,
 	.opt_set_funcs = opt_set_funcs,
+	.verify = VERIFY_OPTIONAL,
 	.help = help
 };
