@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2013-2021 Canonical, Ltd.
- * Copyright (C) 2022-2023 Colin Ian King.
+ * Copyright (C) 2022-2024 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -18,6 +18,9 @@
  *
  */
 #include "stress-ng.h"
+#include "core-affinity.h"
+#include "core-builtin.h"
+#include "core-killpid.h"
 
 #if defined(HAVE_MQUEUE_H)
 #include <mqueue.h>
@@ -39,10 +42,13 @@ static const stress_help_t help[] = {
 	{ NULL, NULL, 			NULL }
 };
 
+typedef int (*switch_func_t)(stress_args_t *args,
+			     const uint64_t switch_freq,
+			     const uint64_t switch_delay,
+			     const uint64_t threshold);
 typedef struct {
 	const char *name;
-	int (*func)(const stress_args_t *args, const uint64_t switch_freq,
-		    const uint64_t switch_delay, const uint64_t threshold);
+	const switch_func_t switch_func;
 } stress_switch_method_t;
 
 #define THRESH_FREQ	(100)		/* Delay adjustment rate in HZ */
@@ -65,15 +71,18 @@ static int stress_set_switch_freq(const char *opt)
  *	report context switch duration
  */
 static void stress_switch_rate(
-	const stress_args_t *args,
-	char *method,
+	stress_args_t *args,
+	const char *method,
 	const double t_start,
 	const double t_end,
 	uint64_t counter)
 {
-	pr_inf("%s: (%s) %.2f nanosecs per context switch (based on parent run time)\n",
-		args->name, method,
-		((t_end - t_start) * STRESS_NANOSECOND) / (double)counter);
+	char msg[128];
+
+	(void)snprintf(msg, sizeof(msg), "nanosecs per context switch (%s method)", method);
+	stress_metrics_set(args, 0, msg,
+		((t_end - t_start) * STRESS_NANOSECOND) / (double)counter,
+		STRESS_HARMONIC_MEAN);
 }
 
 /*
@@ -82,7 +91,7 @@ static void stress_switch_rate(
  *	a specified frequency
  */
 static void stress_switch_delay(
-	const stress_args_t *args,
+	stress_args_t *args,
 	const uint64_t switch_delay,
 	const uint64_t threshold,
 	const double t_start,
@@ -102,7 +111,7 @@ static void stress_switch_delay(
 	 */
 	if (++i >= threshold) {
 		double overrun, overrun_by, t;
-		const uint64_t counter = get_counter(args);
+		const uint64_t counter = stress_bogo_get(args);
 
 		i = 0;
 		t = t_start + ((double)(counter * switch_delay) / STRESS_NANOSECOND);
@@ -129,17 +138,20 @@ static void stress_switch_delay(
  *	synchronization method
  */
 static int stress_switch_pipe(
-	const stress_args_t *args,
+	stress_args_t *args,
 	const uint64_t switch_freq,
 	const uint64_t switch_delay,
 	const uint64_t threshold)
 {
 	pid_t pid;
-	int pipefds[2];
+	int pipefds[2], parent_cpu;
 	size_t buf_size;
 	char *buf;
 
-	(void)memset(pipefds, 0, sizeof(pipefds));
+	if (stress_sig_stop_stressing(args->name, SIGPIPE) < 0)
+		return EXIT_FAILURE;
+
+	(void)shim_memset(pipefds, 0, sizeof(pipefds));
 #if defined(HAVE_PIPE2) &&	\
     defined(O_DIRECT)
 	if (pipe2(pipefds, O_DIRECT) < 0) {
@@ -162,8 +174,10 @@ static int stress_switch_pipe(
 	buf_size = args->page_size;
 #endif
 
-	buf = malloc(buf_size);
-	if (!buf) {
+	buf = (char *)stress_mmap_populate(NULL, buf_size,
+			PROT_READ | PROT_WRITE,
+			MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);
+	if (buf == MAP_FAILED) {
 		pr_fail("%s: pipe read/write buffer allocation failed\n",
 			args->name);
 		(void)close(pipefds[0]);
@@ -174,48 +188,52 @@ static int stress_switch_pipe(
 #if defined(F_SETPIPE_SZ)
 	if (fcntl(pipefds[0], F_SETPIPE_SZ, buf_size) < 0) {
 		pr_dbg("%s: could not force pipe size to 1 page, "
-			"errno = %d (%s)\n",
+			"errno=%d (%s)\n",
 			args->name, errno, strerror(errno));
 	}
 	if (fcntl(pipefds[1], F_SETPIPE_SZ, buf_size) < 0) {
 		pr_dbg("%s: could not force pipe size to 1 page, "
-			"errno = %d (%s)\n",
+			"errno=%d (%s)\n",
 			args->name, errno, strerror(errno));
 	}
 #endif
 
 	stress_set_proc_state(args->name, STRESS_STATE_RUN);
 again:
+	parent_cpu = stress_get_cpu();
 	pid = fork();
 	if (pid < 0) {
-		if (stress_redo_fork(errno))
+		if (stress_redo_fork(args, errno))
 			goto again;
 		(void)close(pipefds[0]);
 		(void)close(pipefds[1]);
-		if (!keep_stressing(args))
+		if (!stress_continue(args))
 			goto finish;
 		pr_fail("%s: fork failed, errno=%d (%s)\n",
 			args->name, errno, strerror(errno));
-		free(buf);
+		(void)munmap((void *)buf, buf_size);
 		return EXIT_FAILURE;
 	} else if (pid == 0) {
+		register const int fd = pipefds[0];
+
+		(void)stress_change_cpu(args, parent_cpu);
 		stress_parent_died_alarm();
 		(void)sched_settings_apply(true);
 
 		(void)close(pipefds[1]);
 
-		while (keep_stressing_flag()) {
+		while (stress_continue_flag()) {
 			ssize_t ret;
 
-			ret = read(pipefds[0], buf, buf_size);
-			if (ret <= 0) {
+			ret = read(fd, buf, buf_size);
+			if (UNLIKELY(ret <= 0)) {
 				if (errno == 0)	/* ret == 0 case */
 					break;
 				if ((errno == EAGAIN) || (errno == EINTR))
 					continue;
 				if (errno == EPIPE)
 					break;
-				pr_fail("%s: read failed, errno=%d (%s)\n",
+				pr_dbg("%s: read failed, errno=%d (%s)\n",
 					args->name, errno, strerror(errno));
 				break;
 			}
@@ -223,49 +241,47 @@ again:
 		(void)close(pipefds[0]);
 		_exit(EXIT_SUCCESS);
 	} else {
-		int status;
 		double t_start;
 		uint64_t delay = switch_delay;
+		register const int fd = pipefds[1];
 
 		/* Parent */
 		(void)close(pipefds[0]);
-		(void)memset(buf, '_', buf_size);
+		(void)shim_memset(buf, '_', buf_size);
 
 		t_start = stress_time_now();
 		do {
 			ssize_t ret;
 
-			inc_counter(args);
+			stress_bogo_inc(args);
 
-			ret = write(pipefds[1], buf, buf_size);
-			if (ret <= 0) {
+			ret = write(fd, buf, buf_size);
+			if (UNLIKELY(ret <= 0)) {
 				if ((errno == EAGAIN) || (errno == EINTR))
 					continue;
 				if (errno == EPIPE)
 					break;
 				if (errno) {
-					pr_fail("%s: write failed, errno=%d (%s)\n",
+					pr_dbg("%s: write failed, errno=%d (%s)\n",
 						args->name, errno, strerror(errno));
 					break;
 				}
 				continue;
 			}
 
-			if (switch_freq)
+			if (UNLIKELY(switch_freq))
 				stress_switch_delay(args, switch_delay, threshold, t_start, &delay);
-		} while (keep_stressing(args));
+		} while (stress_continue(args));
 
-		stress_switch_rate(args, "pipe", t_start, stress_time_now(), get_counter(args));
+		stress_switch_rate(args, "pipe", t_start, stress_time_now(), stress_bogo_get(args));
 
-		(void)close(pipefds[0]);
-
-		(void)kill(pid, SIGKILL);
-		(void)shim_waitpid(pid, &status, 0);
+		(void)close(pipefds[1]);
+		(void)stress_kill_pid_wait(pid, NULL);
 	}
 finish:
 	stress_set_proc_state(args->name, STRESS_STATE_DEINIT);
 
-	free(buf);
+	(void)munmap((void *)buf, buf_size);
 
 	return EXIT_SUCCESS;
 }
@@ -278,14 +294,13 @@ finish:
  *	synchronization method
  */
 static int stress_switch_sem_sysv(
-	const stress_args_t *args,
+	stress_args_t *args,
 	const uint64_t switch_freq,
 	const uint64_t switch_delay,
 	const uint64_t threshold)
 {
 	pid_t pid;
-	int sem_id = -1;
-	int i;
+	int i, sem_id = -1, parent_cpu;
 
 	for (i = 0; i < 100; i++) {
 		key_t key_id = (key_t)stress_mwc16();
@@ -302,77 +317,77 @@ static int stress_switch_sem_sysv(
 
 	stress_set_proc_state(args->name, STRESS_STATE_RUN);
 again:
+	parent_cpu = stress_get_cpu();
 	pid = fork();
 	if (pid < 0) {
-		if (stress_redo_fork(errno))
+		if (stress_redo_fork(args, errno))
 			goto again;
-		if (!keep_stressing(args))
+		if (!stress_continue(args))
 			goto finish;
 		pr_fail("%s: fork failed, errno=%d (%s)\n",
 			args->name, errno, strerror(errno));
 		return EXIT_FAILURE;
 	} else if (pid == 0) {
+		(void)stress_change_cpu(args, parent_cpu);
 		stress_parent_died_alarm();
 		(void)sched_settings_apply(true);
 
-		while (keep_stressing_flag()) {
-			struct sembuf sem;
+		while (stress_continue_flag()) {
+			struct sembuf sem ALIGN64;
 
 			sem.sem_num = 0;
 			sem.sem_op = -1;
 			sem.sem_flg = SEM_UNDO;
 
-			if (semop(sem_id, &sem, 1) < 0)
+			if (UNLIKELY(semop(sem_id, &sem, 1) < 0))
 				break;
 
 			sem.sem_num = 0;
 			sem.sem_op = 1;
 			sem.sem_flg = SEM_UNDO;
 
-			if (semop(sem_id, &sem, 1) < 0)
+			if (UNLIKELY(semop(sem_id, &sem, 1) < 0))
 				break;
 		}
 		_exit(EXIT_SUCCESS);
 	} else {
-		int status;
 		double t_start;
 		uint64_t delay = switch_delay;
-		struct sembuf sem;
+		struct sembuf sem ALIGN64;
 
 		/* Parent */
 		t_start = stress_time_now();
 		do {
-			inc_counter(args);
+			stress_bogo_inc(args);
 
 			sem.sem_num = 0;
 			sem.sem_op = 1;
 			sem.sem_flg = SEM_UNDO;
 
-			if (semop(sem_id, &sem, 1) < 0)
+			if (UNLIKELY(semop(sem_id, &sem, 1) < 0))
 				break;
 
-			if (switch_freq)
+			if (UNLIKELY(switch_freq))
 				stress_switch_delay(args, switch_delay, threshold, t_start, &delay);
 
-			if (!keep_stressing(args))
+			if (!stress_continue(args))
 				break;
 			sem.sem_num = 0;
 			sem.sem_op = -1;
 			sem.sem_flg = SEM_UNDO;
 
-			if (semop(sem_id, &sem, 1) < 0)
+			if (UNLIKELY(semop(sem_id, &sem, 1) < 0))
 				break;
-		} while (keep_stressing(args));
+		} while (stress_continue(args));
 
-		stress_switch_rate(args, "sem-sysv", t_start, stress_time_now(), 2 * get_counter(args));
+		stress_switch_rate(args, "sem-sysv", t_start, stress_time_now(), 2 * stress_bogo_get(args));
 
-		(void)kill(pid, SIGKILL);
-		(void)shim_waitpid(pid, &status, 0);
+		(void)stress_kill_pid_wait(pid, NULL);
 	}
 finish:
 	stress_set_proc_state(args->name, STRESS_STATE_DEINIT);
 
-	semctl(sem_id, 0, IPC_RMID);
+	(void)semctl(sem_id, 0, IPC_RMID);
 
 	return EXIT_SUCCESS;
 }
@@ -386,7 +401,7 @@ finish:
  *	stress by heavy context switching using message queue
  */
 static int stress_switch_mq(
-	const stress_args_t *args,
+	stress_args_t *args,
 	const uint64_t switch_freq,
 	const uint64_t switch_delay,
 	const uint64_t threshold)
@@ -400,6 +415,7 @@ static int stress_switch_mq(
 	char mq_name[64];
 	struct mq_attr attr;
 	stress_msg_t msg;
+	int parent_cpu;
 
 	(void)snprintf(mq_name, sizeof(mq_name), "/%s-%" PRIdMAX "-%" PRIu32,
 			args->name, (intmax_t)args->pid, args->instance);
@@ -414,30 +430,31 @@ static int stress_switch_mq(
                 return EXIT_FAILURE;
 	}
 
-	(void)memset(&msg, 0, sizeof(msg));
+	(void)shim_memset(&msg, 0, sizeof(msg));
 	stress_set_proc_state(args->name, STRESS_STATE_RUN);
 again:
+	parent_cpu = stress_get_cpu();
 	pid = fork();
 	if (pid < 0) {
-		if (stress_redo_fork(errno))
+		if (stress_redo_fork(args, errno))
 			goto again;
-		if (!keep_stressing(args))
+		if (!stress_continue(args))
 			goto finish;
 		pr_fail("%s: fork failed, errno=%d (%s)\n",
 			args->name, errno, strerror(errno));
 		return EXIT_FAILURE;
 	} else if (pid == 0) {
+		(void)stress_change_cpu(args, parent_cpu);
 		stress_parent_died_alarm();
 		(void)sched_settings_apply(true);
 
-		while (keep_stressing_flag()) {
+		while (LIKELY(stress_continue_flag())) {
 			msg.value++;
-			if (mq_send(mq, (char *)&msg, sizeof(msg), 0) < 0)
+			if (UNLIKELY(mq_send(mq, (char *)&msg, sizeof(msg), 0) < 0))
 				break;
 		}
 		_exit(EXIT_SUCCESS);
 	} else {
-		int status;
 		double t_start;
 		uint64_t delay = switch_delay;
 
@@ -446,18 +463,17 @@ again:
 		do {
 			unsigned int prio;
 
-			inc_counter(args);
-			if (mq_receive(mq, (char *)&msg, sizeof(msg), &prio) < 0)
+			stress_bogo_inc(args);
+			if (UNLIKELY(mq_receive(mq, (char *)&msg, sizeof(msg), &prio) < 0))
 				break;
 
-			if (switch_freq)
+			if (UNLIKELY(switch_freq))
 				stress_switch_delay(args, switch_delay, threshold, t_start, &delay);
-		} while (keep_stressing(args));
+		} while (stress_continue(args));
 
-		stress_switch_rate(args, "mq", t_start, stress_time_now(), get_counter(args));
+		stress_switch_rate(args, "mq", t_start, stress_time_now(), stress_bogo_get(args));
 
-		(void)kill(pid, SIGKILL);
-		(void)shim_waitpid(pid, &status, 0);
+		(void)stress_kill_pid_wait(pid, NULL);
 	}
 finish:
 	stress_set_proc_state(args->name, STRESS_STATE_DEINIT);
@@ -476,10 +492,10 @@ static stress_switch_method_t stress_switch_methods[] = {
 	{ "mq",		stress_switch_mq },
 #endif
 	{ "pipe",	stress_switch_pipe },
-#if defined(HAVE_SEM_SYSV)
+#if defined(HAVE_SEM_SYSV) &&	\
+    defined(HAVE_KEY_T)
 	{ "sem-sysv",	stress_switch_sem_sysv },
 #endif
-	{ NULL,		NULL },
 };
 
 /*
@@ -488,18 +504,18 @@ static stress_switch_method_t stress_switch_methods[] = {
  */
 static int stress_set_switch_method(const char *name)
 {
-	stress_switch_method_t *info;
+	size_t i;
 
-	for (info = stress_switch_methods; info->name; info++) {
-		if (!strcmp(info->name, name)) {
-			stress_set_setting("switch-method", TYPE_ID_UINTPTR_T, &info);
+	for (i = 0; i < SIZEOF_ARRAY(stress_switch_methods); i++) {
+		if (!strcmp(stress_switch_methods[i].name, name)) {
+			stress_set_setting("switch-method", TYPE_ID_SIZE_T, &i);
 			return 0;
 		}
 	}
 
 	(void)fprintf(stderr, "switch-method must be one of:");
-	for (info = stress_switch_methods; info->name; info++) {
-		(void)fprintf(stderr, " %s", info->name);
+	for (i = 0; i < SIZEOF_ARRAY(stress_switch_methods); i++) {
+		(void)fprintf(stderr, " %s", stress_switch_methods[i].name);
 	}
 	(void)fprintf(stderr, "\n");
 
@@ -510,19 +526,18 @@ static int stress_set_switch_method(const char *name)
  *  stress_switch
  *	stress by heavy context switching
  */
-static int stress_switch(const stress_args_t *args)
+static int stress_switch(stress_args_t *args)
 {
 	uint64_t switch_freq = 0, switch_delay, threshold;
-
-	stress_switch_method_t *switch_method;
+	size_t switch_method;
 
 	(void)stress_get_setting("switch-freq", &switch_freq);
-	(void)stress_get_setting("switch-method", (void *)&switch_method);
+	(void)stress_get_setting("switch-method", &switch_method);
 
 	switch_delay = (switch_freq == 0) ? 0 : STRESS_NANOSECOND / switch_freq;
 	threshold = switch_freq / THRESH_FREQ;
 
-	return switch_method->func(args, switch_freq, switch_delay, threshold);
+	return stress_switch_methods[switch_method].switch_func(args, switch_freq, switch_delay, threshold);
 }
 
 static void stress_switch_set_default(void)
