@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2013-2021 Canonical, Ltd.
- * Copyright (C)      2022 Colin Ian King.
+ * Copyright (C) 2022-2024 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -18,7 +18,10 @@
  *
  */
 #include "stress-ng.h"
+#include "core-affinity.h"
+#include "core-builtin.h"
 #include "core-cpu.h"
+#include "core-killpid.h"
 #include "core-net.h"
 
 #if defined(HAVE_LINUX_SOCKIOS_H)
@@ -107,8 +110,8 @@ static int stress_set_udp_if(const char *name)
 	return stress_set_setting("udp-if", TYPE_ID_STR, name);
 }
 
-static int stress_udp_client(
-	const stress_args_t *args,
+static int OPTIMIZE3 stress_udp_client(
+	stress_args_t *args,
 	const pid_t mypid,
 	const int udp_domain,
 	const int udp_proto,
@@ -118,19 +121,21 @@ static int stress_udp_client(
 {
 	struct sockaddr *addr = NULL;
 	int rc = EXIT_FAILURE;
+	const pid_t pid = getpid();
 
 	stress_parent_died_alarm();
 	(void)sched_settings_apply(true);
 
 	do {
-		char buf[UDP_BUF];
 		socklen_t len;
-		int fd;
-		int j = 0;
+		int fd, j = 0;
+		char ALIGN64 buf[UDP_BUF];
+		pid_t *pidptr = (pid_t *)buf;
 
 		if ((fd = socket(udp_domain, SOCK_DGRAM, udp_proto)) < 0) {
 			pr_fail("%s: socket failed, errno=%d (%s)\n",
 				args->name, errno, strerror(errno));
+			rc = EXIT_NO_RESOURCE;
 			goto child_die;
 		}
 
@@ -138,6 +143,7 @@ static int stress_udp_client(
 				udp_domain, udp_port, udp_if,
 				&addr, &len, NET_ADDR_ANY) < 0) {
 			(void)close(fd);
+			rc = EXIT_NO_RESOURCE;
 			goto child_die;
 		}
 #if defined(IPPROTO_UDPLITE) &&	\
@@ -147,10 +153,11 @@ static int stress_udp_client(
 			socklen_t slen;
 
 			slen = sizeof(val);
-			if (setsockopt(fd, SOL_UDPLITE, UDPLITE_SEND_CSCOV, &val, slen) < 0) {
+			if (UNLIKELY(setsockopt(fd, SOL_UDPLITE, UDPLITE_SEND_CSCOV, &val, slen) < 0)) {
 				pr_fail("%s: setsockopt failed, errno=%d (%s)\n",
 					args->name, errno, strerror(errno));
 				(void)close(fd);
+				rc = EXIT_NO_RESOURCE;
 				goto child_die;
 			}
 			slen = sizeof(val);
@@ -185,7 +192,7 @@ static int stress_udp_client(
 			socklen_t slen = sizeof(val);
 
 			ret = getsockopt(fd, udp_proto, UDP_CORK, &val, &slen);
-			if (ret == 0) {
+			if (LIKELY(ret == 0)) {
 				slen = sizeof(val);
 				VOID_RET(int, setsockopt(fd, udp_proto, UDP_CORK, &val, slen));
 			}
@@ -199,7 +206,7 @@ static int stress_udp_client(
 			socklen_t slen = sizeof(val);
 
 			ret = getsockopt(fd, udp_proto, UDP_ENCAP, &val, &slen);
-			if (ret == 0) {
+			if (LIKELY(ret == 0)) {
 				slen = sizeof(val);
 				VOID_RET(int, setsockopt(fd, udp_proto, UDP_ENCAP, &val, slen));
 			}
@@ -213,7 +220,7 @@ static int stress_udp_client(
 			socklen_t slen = sizeof(val);
 
 			ret = getsockopt(fd, udp_proto, UDP_NO_CHECK6_TX, &val, &slen);
-			if (ret == 0) {
+			if (LIKELY(ret == 0)) {
 				slen = sizeof(val);
 				VOID_RET(int, setsockopt(fd, udp_proto, UDP_NO_CHECK6_TX, &val, slen));
 			}
@@ -226,7 +233,7 @@ static int stress_udp_client(
 			int val, ret;
 			socklen_t slen = sizeof(val);
 			ret = getsockopt(fd, udp_proto, UDP_NO_CHECK6_RX, &val, &slen);
-			if (ret == 0) {
+			if (LIKELY(ret == 0)) {
 				slen = sizeof(val);
 				VOID_RET(int, setsockopt(fd, udp_proto, UDP_NO_CHECK6_RX, &val, slen));
 			}
@@ -240,7 +247,7 @@ static int stress_udp_client(
 			socklen_t slen = sizeof(val);
 
 			ret = getsockopt(fd, udp_proto, UDP_SEGMENT, &val, &slen);
-			if (ret == 0) {
+			if (LIKELY(ret == 0)) {
 				slen = sizeof(val);
 				VOID_RET(int, setsockopt(fd, udp_proto, UDP_SEGMENT, &val, slen));
 			}
@@ -248,20 +255,32 @@ static int stress_udp_client(
 #else
 		UNEXPECTED
 #endif
+		(void)shim_memset(buf, stress_mwc8(), sizeof(buf));
+		*pidptr = pid;
+
 		do {
-			size_t i;
+			register size_t i;
 
 			for (i = 16; i < sizeof(buf); i += 16, j++) {
 				ssize_t ret;
 
-				(void)memset(buf, 'A' + (j % 26), sizeof(buf));
 				ret = sendto(fd, buf, i, 0, addr, len);
-				if (ret < 0) {
+				if (UNLIKELY(ret < 0)) {
 					if ((errno == EINTR) || (errno == ENETUNREACH))
 						break;
-					pr_fail("%s: sendto failed, errno=%d (%s)\n",
-						args->name, errno, strerror(errno));
-					break;
+					if ((errno == ENOBUFS) || (errno == ENOMEM)) {
+						(void)shim_usleep(10000);
+						continue;
+					}
+					if (errno == EPERM) {
+						(void)shim_usleep(250000);
+						continue;
+					}
+					pr_fail("%s: sendto on port %d failed, errno=%d (%s)\n",
+						args->name, udp_port, errno, strerror(errno));
+					rc = EXIT_FAILURE;
+					(void)close(fd);
+					goto child_die;
 				}
 			}
 #if defined(SIOCOUTQ)
@@ -273,17 +292,16 @@ static int stress_udp_client(
 #else
 		UNEXPECTED
 #endif
-		} while (keep_stressing(args));
+		} while (stress_continue(args));
 		(void)close(fd);
-	} while (keep_stressing(args));
+	} while (stress_continue(args));
 
 	rc = EXIT_SUCCESS;
 child_die:
-
 #if defined(AF_UNIX) &&		\
     defined(HAVE_SOCKADDR_UN)
 	if ((udp_domain == AF_UNIX) && addr) {
-		struct sockaddr_un *addr_un = (struct sockaddr_un *)addr;
+		const struct sockaddr_un *addr_un = (struct sockaddr_un *)addr;
 
 		(void)shim_unlink(addr_un->sun_path);
 	}
@@ -291,18 +309,18 @@ child_die:
 	return rc;
 }
 
-static int stress_udp_server(
-	const stress_args_t *args,
-	const pid_t pid,
+static int OPTIMIZE3 stress_udp_server(
+	stress_args_t *args,
 	const pid_t mypid,
+	const pid_t client_pid,
 	const int udp_domain,
 	const int udp_proto,
 	const int udp_port,
 	const bool udp_gro,
 	const char *udp_if)
 {
-	char buf[UDP_BUF];
-	int fd, status;
+	char ALIGN64 buf[UDP_BUF];
+	int fd;
 #if !defined(__minix__)
 	int so_reuseaddr = 1;
 #endif
@@ -333,11 +351,6 @@ static int stress_udp_server(
 		}
 	}
 #endif
-	if (bind(fd, addr, addr_len) < 0) {
-		pr_fail("%s: bind failed, errno=%d (%s)\n",
-			args->name, errno, strerror(errno));
-		goto die_close;
-	}
 #if !defined(__minix__)
 	if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &so_reuseaddr, sizeof(so_reuseaddr)) < 0) {
 		/*
@@ -350,6 +363,11 @@ static int stress_udp_server(
 		}
 	}
 #endif
+	if (bind(fd, addr, addr_len) < 0) {
+		pr_fail("%s: bind failed, errno=%d (%s)\n",
+			args->name, errno, strerror(errno));
+		goto die_close;
+	}
 
 #if defined(UDP_GRO)
 	if (udp_gro) {
@@ -375,16 +393,36 @@ static int stress_udp_server(
 		UNEXPECTED
 #endif
 		n = recvfrom(fd, buf, sizeof(buf), 0, addr, &len);
-		if (n == 0)
-			break;
-		if (n < 0) {
-			if (errno != EINTR)
+		if (UNLIKELY(n <= 0)) {
+			if (n == 0)
+				break;
+			if (errno == ENOBUFS) {
+				(void)shim_usleep(10000);
+				continue;
+			}
+			if (errno != EINTR) {
 				pr_fail("%s: recvfrom failed, errno=%d (%s)\n",
 					args->name, errno, strerror(errno));
+				rc = EXIT_FAILURE;
+				goto die_close;
+			}
 			break;
+		} else {
+			const pid_t *pidptr = (const pid_t *)buf;
+			const pid_t pid = *pidptr;
+
+			if (UNLIKELY(pid != client_pid)) {
+				pr_fail("%s: server received unexpected data "
+					"contents, got 0x%" PRIxMAX ", "
+					"expected 0x%" PRIxMAX "\n",
+					args->name, (intmax_t)pid,
+					(intmax_t)client_pid);
+				rc = EXIT_FAILURE;
+				goto die_close;
+			}
+			stress_bogo_inc(args);
 		}
-		inc_counter(args);
-	} while (keep_stressing(args));
+	} while (stress_continue(args));
 
 	rc = EXIT_SUCCESS;
 die_close:
@@ -395,15 +433,11 @@ die:
 #if defined(AF_UNIX) &&		\
     defined(HAVE_SOCKADDR_UN)
 	if ((udp_domain == AF_UNIX) && addr) {
-		struct sockaddr_un *addr_un = (struct sockaddr_un *)addr;
+		const struct sockaddr_un *addr_un = (struct sockaddr_un *)addr;
 
 		(void)shim_unlink(addr_un->sun_path);
 	}
 #endif
-	if (pid) {
-		(void)kill(pid, SIGKILL);
-		(void)shim_waitpid(pid, &status, 0);
-	}
 	return rc;
 }
 
@@ -411,12 +445,12 @@ die:
  *  stress_udp
  *	stress by heavy udp ops
  */
-static int stress_udp(const stress_args_t *args)
+static int stress_udp(stress_args_t *args)
 {
 	int udp_port = DEFAULT_UDP_PORT;
 	int udp_domain = AF_INET;
 	pid_t pid, mypid = getpid();
-	int rc = EXIT_SUCCESS, reserved_port;
+	int rc = EXIT_SUCCESS, reserved_port, parent_cpu;
 	int udp_proto = 0;
 #if defined(IPPROTO_UDPLITE)
 	bool udp_lite = false;
@@ -424,6 +458,9 @@ static int stress_udp(const stress_args_t *args)
 	bool udp_gro = false;
 	char *udp_if = NULL;
 
+	if (stress_sigchld_set_handler(args) < 0)
+		return EXIT_NO_RESOURCE;
+
 	(void)stress_get_setting("udp-if", &udp_if);
 	(void)stress_get_setting("udp-port", &udp_port);
 	(void)stress_get_setting("udp-domain", &udp_domain);
@@ -470,21 +507,26 @@ static int stress_udp(const stress_args_t *args)
 
 	stress_set_proc_state(args->name, STRESS_STATE_RUN);
 again:
+	parent_cpu = stress_get_cpu();
 	pid = fork();
 	if (pid < 0) {
-		if (keep_stressing_flag() && (errno == EAGAIN))
+		if (stress_continue_flag() && (errno == EAGAIN))
 			goto again;
 		pr_fail("%s: fork failed, errno=%d (%s)\n",
 			args->name, errno, strerror(errno));
 		return EXIT_FAILURE;
 	} else if (pid == 0) {
+		(void)stress_change_cpu(args, parent_cpu);
 		rc = stress_udp_client(args, mypid, udp_domain, udp_proto, udp_port, udp_gro, udp_if);
-
-		/* Inform parent we're all done */
-		(void)kill(getppid(), SIGALRM);
 		_exit(rc);
 	} else {
-		rc = stress_udp_server(args, pid, mypid, udp_domain, udp_proto, udp_port, udp_gro, udp_if);
+		int status;
+
+		rc = stress_udp_server(args, mypid, pid, udp_domain, udp_proto, udp_port, udp_gro, udp_if);
+		(void)stress_kill_pid_wait(pid, &status);
+		if (WIFEXITED(status))
+			if (WEXITSTATUS(status) != EXIT_SUCCESS)
+				rc = WEXITSTATUS(status);
 	}
 	return rc;
 }
@@ -502,5 +544,6 @@ stressor_info_t stress_udp_info = {
 	.stressor = stress_udp,
 	.class = CLASS_NETWORK | CLASS_OS,
 	.opt_set_funcs = opt_set_funcs,
+	.verify = VERIFY_ALWAYS,
 	.help = help
 };
